import json
from pathlib import Path
from typing import Optional

import typer

from medicure.cli.base import app
from medicure.cli.utils import create_param_help, get_base_path

save_app = typer.Typer()
app.add_typer(save_app, name='save')


@save_app.command('tmdb-info')
def save_tmdb_info(
    api_key: str = typer.Argument(
        ..., help=create_param_help(': str', 'The TMDB API key')
    ),
) -> None:
    """
    Saves TMDB info on disk.
    """
    info_path = get_base_path() / 'tmdb_info.json'
    info = {
        'api_key': api_key,
    }
    with open(info_path, 'w') as f:
        json.dump(info, f, indent=4, sort_keys=True, ensure_ascii=False)

    typer.secho(
        'Your TMDB info has been saved successfully.',
        fg='green',
    )


@save_app.command('collection-info')
def save_collection_info(
    movies_directory: Optional[Path] = typer.Option(
        None,
        '-md',
        '--movies-directory',
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        allow_dash=True,
        help=create_param_help(
            'Directory of your movies, must be given if',
            '`--tvshows-directory` has been not given.',
            internal=True,
            option=True,
        ),
    ),
    tvshows_directory: Optional[Path] = typer.Option(
        None,
        '-td',
        '--tvshows-directory',
        exists=True,
        file_okay=False,
        dir_okay=True,
        writable=True,
        allow_dash=True,
        help=create_param_help(
            'Directory of your tv shows, must be given if',
            '`--movies-directory` has been not given.',
            internal=True,
            option=True,
        ),
    ),
) -> None:
    """
    Saves collection info on disk.
    """
    if movies_directory is None and tvshows_directory is None:
        typer.secho(
            'Error: At least one of the `--movies-directory` '
            'and `--tvshows-directory` should be given.',
            err=True,
            fg='red',
        )
        raise typer.Exit(code=1)

    info_path = get_base_path() / 'collection_info.json'
    info = {
        'movies_directory': str(movies_directory),
        'tvshows_directory': str(tvshows_directory),
    }
    with open(info_path, 'w') as f:
        json.dump(info, f, indent=4, sort_keys=True, ensure_ascii=False)

    typer.secho(
        'Your collection info has been saved successfully.',
        fg='green',
    )



import json
from typing import Any, Dict, List, Optional, TypeVar, Union

import click

T = TypeVar('T')


class JsonParamType(click.ParamType):
    """
    Represents json type.
    """

    name = 'json'

    def convert(
        self,
        value: Optional[Union[Dict[Any, Any], str, bytes]],
        param: click.Parameter,
        ctx: click.Context,
    ) -> Optional[Dict[Any, Any]]:
        """
        Convert the value to the correct type.
        """
        try:
            if isinstance(value, (str, bytes)):
                return json.loads(value)
            else:
                self.fail(
                    f'{value!r} was not a `str`, `bytes` or `None`.',
                    param,
                    ctx,
                )
        except OSError as e:
            self.fail([*e.args, None][0], param, ctx)
        except json.JSONDecodeError as e:
            self.fail(f'Bad JSON: {[*e.args, None][0]}', param, ctx)


Json = JsonParamType()


class DataList(click.ParamType):
    """
    Represents a list containing dataclass objects.
    """

    name = 'dataclass'

    def __init__(self, dataclass: T) -> None:
        self.dataclass = dataclass

    def convert(
        self,
        value: Optional[
            Union[List[Dict[str, Any]], List[List[Any]], str, bytes]
        ],
        param: click.Parameter,
        ctx: click.Context,
    ) -> Optional[List[T]]:
        """
        Convert the value to the correct type.
        """
        try:
            if isinstance(value, (str, bytes)):
                json_value = json.loads(value)
                if isinstance(json_value[0], list):
                    return [self.dataclass(*data) for data in json_value]
                else:
                    return [self.dataclass(**data) for data in json_value]
            else:
                self.fail(
                    f'{value!r} was not a `str`, `bytes` or `None`.',
                    param,
                    ctx,
                )
        except OSError as e:
            self.fail([*e.args, None][0], param, ctx)
        except json.JSONDecodeError as e:
            self.fail(f'Bad JSON: {[*e.args, None][0]}', param, ctx)



import re
from pathlib import Path


def extract_episode_number(file_name: Path) -> int:
    """
    Extracts episode number from file name
    """
    for pattern in (r'(?<=[xXeE])\d{2}', r'\d{1,2}'):
        match = re.search(pattern, str(file_name))
        if match is not None:
            episode_number = int(match.group())
            return episode_number

    raise ValueError('File name did not match to any pattern.')


def escape_nonpath_characters(string: str) -> str:
    """
    Escapes all non-path characters from a string.
    """
    return string.translate(str.maketrans(':/', '  '))



from typing import Optional

import typer

from medicure.cli.typer_patch import patch_param_type, patched
from medicure.version import __version__

if not patched():
    patch_param_type()

app = typer.Typer(
    name='medicure',
    help='Medicure\'s Command-line Interface',
    context_settings=dict(help_option_names=['-h', '--help']),
)


@app.callback(invoke_without_command=True, no_args_is_help=True)
def version_callback(
    version: Optional[bool] = typer.Option(
        None, '-v', '--version', is_eager=True, help='Show medicure version.'
    )
) -> None:
    """
    Adds --version option to cli.
    """
    if version:
        print(__version__)
        raise typer.Exit()



import json
import platform
from pathlib import Path
from typing import Dict

import typer


def get_base_path() -> Path:
    """
    Get base path for medicure
    """
    base_path = None

    # Windows
    if platform.system() == 'Windows':
        base_path = Path('~\\AppData\\Local\\Programs\\medicure')
    # macOS
    elif platform.system() == 'Darwin':
        base_path = Path('~/Library/Application Support/medicure͏͏͏͏')
    # Linux
    elif platform.system() == 'Linux':
        base_path = Path('~/.config/medicure')

    base_path = base_path.expanduser()
    if not base_path.exists():
        base_path.mkdir()

    return base_path


def load_tmdb_info() -> Dict[str, str]:
    """
    Loads TMDB info from disk
    """
    info_path = get_base_path() / 'tmdb_info.json'
    if not info_path.exists():
        typer.secho('Error: No TMDB info found.', err=True, fg='red')
        typer.secho(
            'You need to save your TMDB info '
            'with `medicure save tmdb-info` command.',
            fg='blue',
        )
        raise typer.Exit(code=1)

    with open(info_path) as f:
        info = json.load(f)

    return info


def load_collection_info() -> Dict[str, Path]:
    """
    Loads collection info from disk.
    """
    info_path = get_base_path() / 'collection_info.json'
    if not info_path.exists():
        typer.secho('Error: No collection info found.', err=True, fg='red')
        typer.secho(
            'Hint: You need to save your collection info '
            'with `medicure save collection-info` command.',
            fg='blue',
        )
        raise typer.Exit(code=1)

    with open(info_path) as f:
        info = json.load(f, object_hook=collection_info_from_json)

    return info


def collection_info_from_json(json_object: Dict[str, str]) -> Dict[str, Path]:
    """
    A from_json function for collection info
    """
    return {k: Path(v) for k, v in json_object.items()}


def create_param_help(
    *lines: str, internal: bool = False, option: bool = False
) -> str:
    """
    Creates a beautiful help message for typer parameter.
    """
    message = '\b\n\b\n{}\n'.format('\n'.join(lines))
    if internal:
        message = message[2:]
    if option:
        message = f'{message}\b\n'
    return message


def create_sub_param_help(*lines: str) -> str:
    """
    Creates a beautiful help message for typer sub-parameter.
    """
    return '\b\n{}\n    {}\n'.format(lines[0], '\n    '.join(lines[1:]))



from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional


@dataclass
class DubbingSupplier:
    """
    Represents a dubbing supplier data structure.

    Attributes
    ----------
    name: str
        Name of dubbing supplier, if dubbing supplier represents
        original audio of movie or tv show should be set to its
        language.

    audio_language_code: str
        Current 3-letter language code for audio track
        See https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes
        for available langauge codes.

    subtitle_language_code: str
        Current 3-letter language code for subtitle track
        See https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes
        for available langauge codes.

    correct_language_code: str
        Correct 3-letter language code for dubbing supplier
        See https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes
        for available langauge codes.

    audio_search_pattern: str, optional
        The search pattern for finding audio track

    subtitle_search_pattern: str, optional
        The search pattern for finding subtitle track

    file_id: int
        The file id which include dubbing supplier tracks
    """

    name: str
    audio_language_code: str
    subtitle_language_code: str
    correct_language_code: str
    audio_search_pattern: Optional[str]
    subtitle_search_pattern: Optional[str]
    file_id: int

    _audio_track_id: int = None
    _subtitle_track_id: int = None

    def __getattr__(self, item: str) -> Any:
        if item in ('_has_audio', '_has_subtitle'):
            tt = item.split('_')[-1]
            return self.__getattribute__(f'_{tt}_track_id') is not None

        return self.__getattribute__(item)


@dataclass
class FileInfo:
    """
    Represents a file info data structure.

    Attributes
    ----------
    path: Path
        The file path
    id: int
        The file id
    """

    path: Path
    id: int



from importlib import reload
from inspect import getsourcefile

import typer


def patched() -> bool:
    """
    If Typer is patched by medicure
    """
    return getattr(typer, 'patched_by_medicure', False)


def patch_param_type() -> None:
    """
    Patches Typer with Click's ParamTypes.
    """
    # Patch typer.models and typer.params
    mp_patches = [
        {
            'after': 'autocompletion: Optional[Callable[..., Any]] = None,',
            'insert': 'param_type: Optional[click.ParamType] = None,',
        },
        {
            'after': 'self.autocompletion = autocompletion',
            'insert': 'self.param_type = param_type',
        },
        {
            'after': 'autocompletion=autocompletion,',
            'insert': 'param_type=param_type,',
        },
    ]
    for module in ('models', 'params'):
        module_source_path = getsourcefile(getattr(typer, module))
        with open(module_source_path) as f:
            module_source = ''
            for line in f:
                module_source += line
                for patch in mp_patches:
                    if patch['after'] in line:
                        line_indent = line[: line.index(patch['after'])]
                        module_source += f'{line_indent}{patch["insert"]}\n'
                        break

        with open(module_source_path, 'w') as f:
            f.write(module_source)

        reload(getattr(typer, module))

    # Patch typer.main
    main_patch = {
        'replace': 'if origin is not None:',
        'with': (
            '\tif parameter_info.param_type:\n'
            '\t\tparameter_type = parameter_info.param_type\n'
            '\tif parameter_type is None and origin is not None:\n'
        ),
    }
    main_source_path = getsourcefile(typer.main)
    with open(main_source_path) as f:
        main_source = ''
        for line in f:
            if main_patch['replace'] in line:
                main_source += main_patch['with'].replace('\t', ' ' * 4)
            else:
                main_source += line

    with open(main_source_path, 'w') as f:
        f.write(main_source)

    reload(typer.main)

    # Mark typer as patched
    init_source_path = getsourcefile(typer)
    with open(init_source_path) as f:
        init_source = f.read()
        init_source += '\npatched_by_medicure = True\n'

    with open(init_source_path, 'w') as f:
        f.write(init_source)

    reload(typer)



import re
from pathlib import Path

from setuptools import find_packages, setup

packages = find_packages(exclude=['tests*'])

requirements_path = Path('requirements')
with open(requirements_path / 'main.txt') as f:
    requirements = f.read().split()

extras_requirements = {}
for extra in requirements_path.iterdir():
    if not extra.name.endswith('main.txt'):
        with open(extra) as f:
            extras_requirements[extra.name[:-4]] = f.read().split()

with open('README.md') as f:
    long_description = f.read()

with open(Path('medicure') / 'version.py') as f:
    version = re.search(
        r'__version__\s+=\s+\'(?P<version>.*)\'', f.read()
    ).group('version')

setup(
    name='medicure',
    version=version,
    description=(
        'A cosmetic treatment for your media files: '
        'movies, tv shows and also their subtitles.'
    ),
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/alirezatheh/medicure',
    project_urls={
        'Bug Tracker': 'https://github.com/alirezatheh/medicure/issues',
        'Documentation': 'https://github.com/alirezatheh/medicure',
        'Source Code': 'https://github.com/alirezatheh/medicure',
    },
    author='Alireza Hosseini',
    author_email='alirezatheh@gmail.com',
    packages=packages,
    entry_points={'console_scripts': ['medicure = medicure.cli:app']},
    keywords=[
        'audio',
        'video',
        'videos',
        'media',
        'multimedia',
        'movies',
        'tv',
        'tvshows',
        'tv-shows',
        'series',
        'subtitle',
        'subtitles',
        'audio-processing',
        'video-processing',
    ],
    install_requires=requirements,
    extras_require=extras_requirements,
    python_requires='>=3.8',
    classifiers=[
        'Topic :: Multimedia :: Video',
        'Topic :: Multimedia :: Sound/Audio',
        'Topic :: Other/Nonlisted Topic',
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: OS Independent',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
    ],
)



import os
import re
from collections import defaultdict
from pathlib import Path
from typing import DefaultDict, Dict, List, Optional

import tmdbsimple as tmdb
import typer
from iso639 import languages
from pymediainfo import MediaInfo

from medicure.data_structures import DubbingSupplier, FileInfo
from medicure.utils import escape_nonpath_characters, extract_episode_number


class Medicure:
    """
    The Medicure
    """

    _media_suffix_pattern = r'\.(mkv|m4v|mp4|mka)$'
    _subtitle_suffix_pattern = r'\.(srt|mks|idx|sub)$'

    def __init__(
        self,
        tmdb_api_key: str,
        movies_directory: Optional[Path] = None,
        tvshows_directory: Optional[Path] = None,
    ) -> None:
        """
        Initializes Medicure.

        Parameters
        ----------
        tmdb_api_key: str
            Your TMDB API key

        movies_directory: Path, optional
            Your movies' directory, this should be given for treating a
            movie.

        tvshows_directory: Path, optional
            Your tv shows' directory, this should be given for treating
            a tv show.
        """
        tmdb.API_KEY = tmdb_api_key
        self._movies_directory = movies_directory
        self._tvshows_directory = tvshows_directory
        self._dubbing_suppliers: Optional[List[DubbingSupplier]] = None
        self._movie_file_infos: Optional[List[FileInfo]] = None
        self._season_file_infos: Optional[
            DefaultDict[int, List[FileInfo]]
        ] = None
        self._video_track_id: Optional[int] = None

    def treat_media(
        self,
        imdb_id: str,
        file_search_pattern_to_id: Dict[str, int],
        video_language_code: str,
        video_source: str,
        video_release_format: str,
        dubbing_suppliers: List[DubbingSupplier],
        season_number: Optional[int] = None,
    ) -> None:
        """
        Fixes video source, audio source, file name and language for
        all tracks.

        Parameters
        ----------
        imdb_id: str
            IMDB id

        file_search_pattern_to_id: dict[str, int]
            Dict of patterns for finding files to file ids

        video_language_code: str
            3-letter language code for video track
            See https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes
            for available langauge codes.

        video_source: str
            Source of the video file; name of encoder or the website
            which video is downloaded from

        video_release_format: str
            Format of the video file eg: Blu-ray, WEBRip, etc. See
            https://en.wikipedia.org/wiki/Pirated_movie_release_types
            for available formats.

        dubbing_suppliers: list[DubbingSupplier]
            List of possible dubbing suppliers

        season_number: int, optional
            If imdb_id is a tv show, season number should be given.
        """
        self._dubbing_suppliers = dubbing_suppliers

        # Getting info from TMDB
        find = tmdb.Find(imdb_id)
        info = find.info(external_source='imdb_id')

        # If id is a movie
        if info['movie_results']:
            assert (
                self._movies_directory is not None
            ), 'Movies directory has been not given for a movie.'
            movie = info['movie_results'][0]
            title = escape_nonpath_characters(movie['title'])
            release_year = movie['release_date'][:5]
            movie_directory = self._movies_directory.joinpath(
                f'{title} - {release_year}'
            )
            self._scan_directory(
                movie_directory,
                file_search_pattern_to_id,
                'movie',
                self._media_suffix_pattern,
            )
            self._reset_tracks_info()
            for file_info in self._movie_file_infos:
                self._save_file_tracks_info(file_info)

            os.system(
                'mkvmerge -o "{output}" {track_config}'
                ''.format(
                    output=movie_directory.joinpath(
                        f'{title} - {release_year}.mkv'
                    ),
                    track_config=self._get_track_config(
                        'movie',
                        video_language_code,
                        video_source,
                        video_release_format,
                    ),
                )
            )

        # If id is a tv show
        elif info['tv_results']:
            assert (
                self._tvshows_directory is not None
            ), 'tv shows directory has been not given for a tv show.'
            assert (
                season_number is not None
            ), '`season_number` has not been given for a tv show.'

            tvshow = info['tv_results'][0]
            name = escape_nonpath_characters(tvshow['name'])
            tmdb_id = tvshow['id']
            season = tmdb.TV_Seasons(tmdb_id, season_number).info()
            season_name = escape_nonpath_characters(season['name'])

            season_directory = self._tvshows_directory.joinpath(
                name, season_name
            )
            self._scan_directory(
                season_directory,
                file_search_pattern_to_id,
                'season',
                self._media_suffix_pattern,
            )

            # Create destination directory if not exists already
            destination_directory = Path(f'{str(season_directory)} Edited')
            if not destination_directory.exists():
                destination_directory.mkdir()

            for episode in season['episodes']:
                self._reset_tracks_info()
                ename = escape_nonpath_characters(episode['name'])
                enumber = episode['episode_number']

                # If episode file does not exist
                if enumber not in self._season_file_infos:
                    continue

                for file_info in self._season_file_infos[enumber]:
                    self._save_file_tracks_info(file_info)

                os.system(
                    'mkvmerge -o "{output}" {track_config}'
                    ''.format(
                        output=destination_directory.joinpath(
                            f'{name} '
                            f'- S{season_number:02d}E{enumber:02d} '
                            f'- {ename}.mkv',
                        ),
                        track_config=self._get_track_config(
                            'tvshow',
                            video_language_code,
                            video_source,
                            video_release_format,
                            enumber,
                        ),
                    )
                )

    def treat_subtitle(
        self,
        imdb_id: str,
        file_search_pattern_to_id: Dict[str, int],
        language_code: str,
        source: Optional[str] = None,
        release_format: Optional[str] = None,
        include_full_information: bool = False,
        season_number: Optional[int] = None,
    ) -> None:
        """
        Fixes subtitle source, file name and language.

        Parameters
        ----------
        imdb_id: str
            IMDB id

        file_search_pattern_to_id: dict[str, int]
            Dict of patterns for finding files to file ids

        language_code: str
            3-letter language code for subtitle
            See https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes
            for available langauge codes.

        source: str, optional
            Source of the subtitle file: name of author or the website
            which subtitle is downloaded from. This should be given only
            when `include_full_information` is `True`.

        release_format: str, optional
            Format of the video that the subtitle is sync with eg:
            Blu-ray, WEBRip, etc. See
            https://en.wikipedia.org/wiki/Pirated_movie_release_types
            for available formats. This should be given only
            when `include_full_information` is `True`.

        include_full_information: bool
            If set to `True` the subtitle will be converted to mks
            format inorder to save all subtitle information. If set to
            `True`, `source` and `release_format`
            should also be given.

        season_number: int, optional
            If imdb_id is a tv show season number should be given.
        """
        # Getting info from TMDB
        find = tmdb.Find(imdb_id)
        info = find.info(external_source='imdb_id')

        track_config = (
            '--language 0:{lc} '
            '--track-name 0:"{tn}" '
            '--sub-charset 0:WINDOWS-1256 '
            '--default-track 0 '
            '--forced-track 0:0 '
        ).format(
            lc=language_code,
            tn=f'{source} {release_format}',
        )

        # If id is a movie
        if info['movie_results']:
            assert (
                self._movies_directory is not None
            ), 'Movies directory has been not given for a movie.'
            movie = info['movie_results'][0]
            title = escape_nonpath_characters(movie['title'])
            release_year = movie['release_date'][:5]
            movie_directory = self._movies_directory.joinpath(
                f'{title} - {release_year}',
            )
            self._scan_directory(
                movie_directory,
                file_search_pattern_to_id,
                'movie',
                self._subtitle_suffix_pattern,
            )
            output = movie_directory.joinpath(
                f'{title} - {release_year}.{language_code}',
            )

            if include_full_information:
                os.system(
                    'mkvmerge -o "{output}" {track_config}"{input}"'
                    ''.format(
                        output=f'{output}.mks',
                        track_config=track_config,
                        input=self._movie_file_infos[0].path,
                    )
                )
            else:
                for file_info in self._movie_file_infos:
                    original_file_path = file_info.path
                    final_file_path = '{output}{ext}'.format(
                        output=output,
                        ext=original_file_path.suffix,
                    )
                    original_file_path.rename(final_file_path)
                    typer.echo(
                        f'The file: {original_file_path} '
                        f'renamed to: {final_file_path}',
                    )

        # If id is a tv show
        elif info['tv_results']:
            assert (
                self._tvshows_directory is not None
            ), 'tv shows directory has been not given for a tv show.'
            assert (
                season_number is not None
            ), '`season_number` has not been given for a tv show.'

            tvshow = info['tv_results'][0]
            name = escape_nonpath_characters(tvshow['name'])
            tmdb_id = tvshow['id']
            season = tmdb.TV_Seasons(tmdb_id, season_number).info()
            season_name = escape_nonpath_characters(season['name'])

            season_directory = self._tvshows_directory.joinpath(
                name, season_name
            )
            self._scan_directory(
                season_directory,
                file_search_pattern_to_id,
                'season',
                self._subtitle_suffix_pattern,
            )

            for episode in season['episodes']:
                ename = escape_nonpath_characters(episode['name'])
                enumber = episode['episode_number']

                # If episode file exists
                if enumber not in self._season_file_infos:
                    continue

                output = season_directory.joinpath(
                    f'{name} - S{season_number:02d}E{enumber:02d} - {ename}'
                    f'.{language_code}',
                )

                if include_full_information:
                    os.system(
                        'mkvmerge -o "{output}" {track_config}"{input}"'
                        ''.format(
                            output=f'{output}.mks',
                            track_config=track_config,
                            input=self._season_file_infos[enumber][0],
                        )
                    )
                else:
                    for file_info in self._season_file_infos[enumber]:
                        original_file_path = file_info.path
                        final_file_path = '{output}{ext}'.format(
                            output=output,
                            ext=original_file_path.suffix,
                        )
                        original_file_path.rename(final_file_path)
                        typer.echo(
                            f'The file: {original_file_path} '
                            f'renamed to: {final_file_path}',
                        )

    def _scan_directory(
        self,
        directory: Path,
        file_search_pattern_to_id: Dict[str, int],
        directory_type: str,
        file_suffix_pattern: str,
    ) -> None:
        setattr(
            self,
            f'_{directory_type}_file_infos',
            list() if directory_type == 'movie' else defaultdict(list),
        )

        for file_name in directory.iterdir():
            if re.search(file_suffix_pattern, file_name.suffix) is None:
                continue

            file_infos = getattr(self, f'_{directory_type}_file_infos')
            if directory_type == 'season':
                file_infos = file_infos[extract_episode_number(file_name)]

            for pattern in file_search_pattern_to_id:
                if re.search(pattern, str(file_name)) is None:
                    continue

                file_infos.append(
                    FileInfo(
                        path=directory / file_name,
                        id=file_search_pattern_to_id[pattern],
                    ),
                )
                break

        if directory_type == 'movie':
            self._movie_file_infos.sort(key=lambda file: file.id)
        # season
        else:
            for enumber in self._season_file_infos:
                self._season_file_infos[enumber].sort(key=lambda file: file.id)

    def _save_file_tracks_info(self, file_info: FileInfo) -> None:
        media_info = MediaInfo.parse(file_info.path)
        for track in media_info.tracks:

            if track.track_type == 'Video':
                self._video_track_id = track.track_id - 1
                continue

            track_types = {
                'audio': 'Audio',
                'subtitle': 'Text',
            }
            for tt in track_types:
                if track.track_type == track_types[tt]:
                    for ds in self._dubbing_suppliers:
                        if (
                            not getattr(ds, f'_has_{tt}')
                            and file_info.id == ds.file_id
                            and (
                                self._match(
                                    getattr(ds, f'{tt}_search_pattern'),
                                    track.title,
                                )
                                or (
                                    ds.audio_language_code is None
                                    and track.language is None
                                )
                                or ds.audio_language_code
                                == languages.get(
                                    part1=track.language,
                                ).part2b
                            )
                        ):
                            setattr(ds, f'_{tt}_track_id', track.track_id - 1)
                    continue

    def _reset_tracks_info(self) -> None:
        for ds in self._dubbing_suppliers:
            ds._audio_track_id = None
            ds._subtitle_track_id = None

    def _get_track_config(
        self,
        id_type: str,
        video_language_code: str,
        video_source: str,
        video_release_format: str,
        episode_number: int = None,
    ) -> str:
        track_config = '--track-order {track_order} '.format(
            track_order=self._get_track_order(),
        )
        file_infos = []
        if id_type == 'movie':
            file_infos = self._movie_file_infos
        elif id_type == 'tvshow':
            file_infos = self._season_file_infos[episode_number]

        for file_info in file_infos:
            file_tracks_ids = {
                'audio': [],
                'subtitle': [],
            }

            if file_info.id == 0:
                track_config += '--language {tid}:{lc} '.format(
                    tid=self._video_track_id,
                    lc=video_language_code,
                )
                track_config += '--default-track {tid} '.format(
                    tid=self._video_track_id,
                )
                track_config += '--forced-track {tid}:0 '.format(
                    tid=self._video_track_id
                )
                track_config += '--track-name {tid}:"{tn}" '.format(
                    tid=self._video_track_id,
                    tn=f'{video_source} {video_release_format}',
                )

            for ds in self._dubbing_suppliers:
                if ds.file_id == file_info.id:
                    for tt in ['audio', 'subtitle']:
                        if getattr(ds, f'_has_{tt}'):
                            track_config += self._make_language_code(ds, tt)
                            track_config += self._make_default_or_forced(
                                ds, tt, 'default-track'
                            )
                            track_config += self._make_default_or_forced(
                                ds, tt, 'forced-track'
                            )
                            track_config += self._make_name(ds, tt)
                            file_tracks_ids[tt].append(
                                getattr(ds, f'_{tt}_track_id'),
                            )

            for tt in ['audio', 'subtitle']:
                if len(file_tracks_ids[tt]) > 0:
                    track_config += '--{tt}-tracks {ts} '.format(
                        tt=tt,
                        ts=','.join([str(tid) for tid in file_tracks_ids[tt]]),
                    )
                else:
                    track_config += f'--no-{tt} '

            track_config += f'"{file_info.path}" '

        return track_config

    def _get_track_order(self) -> str:
        track_order = f'0:{self._video_track_id}'

        for tt in ['audio', 'subtitle']:
            for ds in self._dubbing_suppliers:
                if getattr(ds, f'_has_{tt}'):
                    track_order += self._make_order(ds, tt)

        return track_order

    @staticmethod
    def _make_order(ds: DubbingSupplier, sa: str) -> str:
        return ',{fid}:{tid}'.format(
            fid=ds.file_id,
            tid=getattr(ds, f'_{sa}_track_id'),
        )

    def _make_name(self, ds: DubbingSupplier, sa: str) -> str:
        return '--track-name {tid}:"{tn}" '.format(
            tid=getattr(ds, f'_{sa}_track_id'),
            tn=ds.name if self._dubbing_suppliers[0].name != ds.name else '',
        )

    def _make_default_or_forced(
        self,
        ds: DubbingSupplier,
        sa: str,
        df: str,
    ) -> str:
        flag = 0
        if df == 'default-track' and sa == 'audio':
            if self._dubbing_suppliers[0].name == ds.name:
                flag = 1
        return '--{df} {tid}:{flag} '.format(
            df=df, tid=getattr(ds, f'_{sa}_track_id'), flag=flag
        )

    @staticmethod
    def _make_language_code(ds: DubbingSupplier, sa: str) -> str:
        return '--language {tid}:{lc} '.format(
            tid=getattr(ds, f'_{sa}_track_id'), lc=ds.correct_language_code
        )

    @staticmethod
    def _match(pattern: str, string: str) -> bool:
        if pattern == r'':
            return True
        if pattern is None or string is None:
            return False
        return re.search(pattern, string) is not None



from typing import Dict, List, Optional

import typer

from medicure.cli.base import app
from medicure.cli.types import DataList, Json
from medicure.cli.utils import (
    create_param_help,
    create_sub_param_help,
    load_collection_info,
    load_tmdb_info,
)
from medicure.core import Medicure
from medicure.data_structures import DubbingSupplier

treat_app = typer.Typer()
app.add_typer(treat_app, name='treat')


@treat_app.command('media')
def treat_media(
    imdb_id: str = typer.Argument(
        ..., help=create_param_help(': str', 'IMDB id')
    ),
    file_search_pattern_to_id: Dict[str, int] = typer.Argument(
        ...,
        param_type=Json,
        help=create_param_help(
            ': dict[str, int]',
            'Dict of patterns for finding files to file ids',
            'You can pass a json like string for this argument.',
        ),
    ),
    video_language_code: str = typer.Argument(
        ...,
        help=create_param_help(
            ': str',
            '3-letter language code for video track',
            'See https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes',
            'for available langauge codes.',
        ),
    ),
    video_source: str = typer.Argument(
        ...,
        help=create_param_help(
            ': str',
            'Source of the video file; name of encoder or the website',
            'which video is downloaded from',
        ),
    ),
    video_release_format: str = typer.Argument(
        ...,
        help=create_param_help(
            ': str',
            'Format of the video file eg: Blu-ray, WEBRip, ...',
            'See https://en.wikipedia.org/wiki/Pirated_movie_release_types',
            'for available formats.',
        ),
    ),
    dubbing_suppliers: List[DubbingSupplier] = typer.Argument(
        ...,
        param_type=DataList(DubbingSupplier),
        help=create_param_help(
            ': list[DubbingSupplier]',
            'List of possible dubbing suppliers ',
            '`DubbingSupplier` is a dataclass which has the following',
            'attributes:',
            create_sub_param_help(
                'name: str',
                'Name of dubbing supplier, if dubbing supplier represents',
                'original audio of movie or tv show should be set to its',
                'language.',
            ),
            create_sub_param_help(
                'audio_language_code: str',
                'Current 3-letter language code for audio track',
            ),
            create_sub_param_help(
                'subtitle_language_code: str',
                'Current 3-letter language code for subtitle track',
            ),
            create_sub_param_help(
                'correct_language_code: str',
                'Correct 3-letter language code for dubbing supplier',
            ),
            create_sub_param_help(
                'audio_search_pattern: str, optional',
                'The search pattern for finding audio track',
            ),
            create_sub_param_help(
                'subtitle_search_pattern: str, optional',
                'The search pattern for finding subtitle track',
            ),
            create_sub_param_help(
                'file_id: str',
                'The file id which include dubbing supplier tracks',
            ),
            create_param_help(
                'You can pass a json like string containing either',
                'list of value lists or list of dict of keyword arguments',
                'for this argument.',
                internal=True,
            ),
        ),
    ),
    season_number: Optional[int] = typer.Argument(
        None,
        help=create_param_help(
            ': int',
            'If `IMDB_ID` is a tv show, season number should be given.',
        ),
    ),
) -> None:
    """
    Fixes video source, audio source, file name and language for
    all tracks.
    """
    medicure = Medicure(load_tmdb_info()['api_key'], **load_collection_info())
    try:
        medicure.treat_media(
            imdb_id,
            file_search_pattern_to_id,
            video_language_code,
            video_source,
            video_release_format,
            dubbing_suppliers,
            season_number,
        )
    except AssertionError as e:
        typer.secho(
            f'Error: `{str(e)[:15].upper()}{str(e)[15:]}', err=True, fg='red'
        )


@treat_app.command('subtitle')
def treat_subtitle(
    imdb_id: str = typer.Argument(
        ..., help=create_param_help(': str', 'IMDB id')
    ),
    file_search_pattern_to_id: Dict[str, int] = typer.Argument(
        ...,
        param_type=Json,
        help=create_param_help(
            ': dict[str, int]',
            'Dict of patterns for finding files to file ids',
            'You can pass a json like string for this argument.',
        ),
    ),
    language_code: str = typer.Argument(
        ...,
        help=create_param_help(
            ': str',
            '3-letter language code for subtitle',
            'See https://en.wikipedia.org/wiki/List_of_ISO_639-2_codes',
            'for available langauge codes.',
        ),
    ),
    source: Optional[str] = typer.Argument(
        None,
        help=create_param_help(
            ': str',
            'Source of the subtitle file: name of author or the website',
            'which subtitle is downloaded from. This should be given only',
            'when `--include-full-information` flag is set.',
        ),
    ),
    release_format: Optional[str] = typer.Argument(
        None,
        help=create_param_help(
            ': str',
            'Format of the video that the subtitle is sync with eg:',
            'Blu-ray, WEBRip, ...',
            'See https://en.wikipedia.org/wiki/Pirated_movie_release_types',
            'for available formats. This should be given only',
            'when `--include-full-information` flag is set.',
        ),
    ),
    include_full_information: bool = typer.Option(
        False,
        '-ifi',
        '--include-full-information',
        help=create_param_help(
            'If this flag is set, the subtitle will be converted to mks',
            'format inorder to save all subtitle information. If set,',
            '`SOURCE` and `RELEASE_FORMAT` should also be given.',
            internal=True,
            option=True,
        ),
    ),
    season_number: Optional[int] = typer.Argument(
        None,
        help=create_param_help(
            ': int',
            'If `IMDB_ID` is a tv show, season number should be given.',
        ),
    ),
) -> None:
    """
    Fixes subtitle source, file name and language.
    """
    subcure = Medicure(load_tmdb_info()['api_key'], **load_collection_info())
    try:
        subcure.treat_subtitle(
            imdb_id,
            file_search_pattern_to_id,
            language_code,
            source,
            release_format,
            include_full_information,
            season_number,
        )
    except AssertionError as e:
        typer.secho(
            f'Error: `{str(e)[:15].upper()}{str(e)[15:]}', err=True, fg='red'
        )
